/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.
 */

#include "host/pingerfactory.h"

#include <cstdlib>

#include <iostream>

#include <boost/asio/ip/tcp_raw_protocol.hpp>
#include <boost/system/system_error.hpp>

#include <logfunc.hpp>

#include "boost_assert_handler.h"
#include "icmp/icmppaketdistributor.h"
#include "icmp/icmppinger.h"
#include "tcp/tcppinger.h"

using namespace std;
using boost::shared_ptr;
using boost::asio::ip::icmp;
using boost::asio::ip::tcp_raw_protocol;
using boost::system::system_error;
using I2n::Logger::GlobalLogger;

//-----------------------------------------------------------------------------
// PingerFactory
//-----------------------------------------------------------------------------

/**
 * @brief Default constructor.
 */
PingerFactory::PingerFactory()
{
}

/**
 * @brief Destructor.
 */
PingerFactory::~PingerFactory()
{
}

/**
 * @brief Create a Pinger object suitable to the given protocol.
 *
 * @param protocol One of the available ping protocols.
 * @param io_serv The one @c io_service object that controls async processing
 * @param network_interface The network interface name from where the ping
 * packet will be sent.
 * @return a Pinger object to able to ping using the given protocol.
 */
PingerItem PingerFactory::createPinger(
        const PingProtocol protocol,
        const IoServiceItem io_serv,
        const string &network_interface,
        const int ping_reply_timeout
)
{
    BOOST_ASSERT( /*( PingProtocol_First <= protocol ) && */( protocol <= PingProtocol_Last ) );
    BOOST_ASSERT( !network_interface.empty() );

    try
    {
        if (protocol == PingProtocol_ICMP)
        {
            PingerItem new_pinger(
                    new IcmpPinger( io_serv, icmp::v4(), network_interface, ping_reply_timeout )
            );
            IcmpPaketDistributor::get_distributor(icmp::v4(), network_interface, io_serv)
                                  ->register_pinger(new_pinger);
            return new_pinger;
        }
        else if (protocol == PingProtocol_ICMPv6)
        {
            PingerItem new_pinger(
                    new IcmpPinger( io_serv, icmp::v6(), network_interface, ping_reply_timeout )
            );
            IcmpPaketDistributor::get_distributor(icmp::v6(), network_interface, io_serv)
                                  ->register_pinger(new_pinger);
            return new_pinger;
        }
        else if (protocol == PingProtocol_TCP)
        {
            return PingerItem(
                    new TcpPinger( io_serv, tcp_raw_protocol::v4(), network_interface, ping_reply_timeout )
            );
        }
        else if (protocol == PingProtocol_TCP_IPv6)
        {
            return PingerItem(
                    new TcpPinger( io_serv, tcp_raw_protocol::v6(), network_interface, ping_reply_timeout )
            );
        }
        else
        {
            BOOST_ASSERT( !"Try to create a pinger from an invalid protocol" );                                                 //lint !e506
            return PingerItem();                                                                                                             //lint !e527
        }
    }
    catch ( const system_error &ex )
    {
        // Raw sockets are locked down by Unix operating systems to prevent
        // malware and security holes, thus it requires root access to use it.
        GlobalLogger.error() << "Pingcheck must run as root (" << ex.what() << ")" << endl;
        exit( EXIT_FAILURE );
    }

    return PingerItem();                                                                                                                        //lint !e527
}

/**
 * @brief Create a Pinger object suitable to the given list of protocols.
 *
 * @param protocol_list A list of the available ping protocols.
 * @param io_serv The @c io_service object.
 * @param network_interface The network interface name from where the ping
 * packet will be sent.
 * @param destination_address The remote address to ping.
 * @param destination_port The remote port to ping.
 * @param nameserver Server to resolve the addresses.
 *
 * @return a Pinger object to able to ping using the given list of protocols.
 */
PingRotateItem PingerFactory::createPinger(
        const PingProtocolList &protocol_list,
        const IoServiceItem io_serv,
        const string &network_interface,
        const string &destination_address,
        const uint16_t destination_port,
        const string &nameserver,
        const int resolved_ip_ttl_threshold,
        const int ping_reply_timeout
)
{
    BOOST_ASSERT( 0 < protocol_list.size() );
    BOOST_ASSERT( !network_interface.empty() );
    BOOST_ASSERT( !destination_address.empty() );
    BOOST_ASSERT( 0 < destination_port );
    BOOST_ASSERT( !nameserver.empty() );

    return PingRotateItem( new PingRotate(
            io_serv,
            network_interface,
            destination_address,
            destination_port,
            nameserver,
            resolved_ip_ttl_threshold,
            ping_reply_timeout,
            protocol_list
    ) );
}
