/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.

 Christian Herdtweck, Intra2net AG 2015
 */

#ifndef ICMP_PAKET_DISTRIBUTOR_H
#define ICMP_PAKET_DISTRIBUTOR_H

#include <set>
#include <map>
#include <boost/shared_ptr.hpp>
#include <boost/asio.hpp>

#include "host/pinger.h"
#include "host/networkinterface.hpp"

using boost::asio::ip::icmp;

// for each IP protocol (v4/v6) and each network interface (string),
//   there can only be one IcmpPaketDistributor instance
typedef std::pair<icmp::socket::protocol_type, std::string> DistributorInstanceIdentifier;

struct IcmpPaketDistributorInstanceIdentifierComparator
{
    bool operator() ( const DistributorInstanceIdentifier &a,
                      const DistributorInstanceIdentifier &b ) const ;
};

class IcmpPaketDistributor;

typedef boost::shared_ptr<IcmpPaketDistributor> IcmpPaketDistributorItem;

//-----------------------------------------------------------------------------
// IcmpPaketDistributor
//-----------------------------------------------------------------------------

class IcmpPaketDistributor
{
public:
    bool register_pinger( const PingerItem new_pinger );
    bool unregister_pinger( const PingerItem old_pinger );

    static IcmpPaketDistributorItem get_distributor(
            const icmp::socket::protocol_type &protocol,
            const std::string &network_interface,
            const IoServiceItem io_serv );

    static IcmpPaketDistributorItem get_distributor(
            const icmp::socket::protocol_type &protocol,
            const std::string &network_interface );

    ~IcmpPaketDistributor() {};

private:
    // hide away constructor, copy constructor and copy operator
    IcmpPaketDistributor(
            const icmp::socket::protocol_type &protocol,
            const std::string &network_interface,
            const IoServiceItem io_serv );
    IcmpPaketDistributor(IcmpPaketDistributor const&);
    void operator=(IcmpPaketDistributor const&);

    void register_receive_handler();
    void handle_receive( const boost::system::error_code &error,
                         const size_t &bytes_transferred );


private:
    /// The socket object
    icmp::socket Socket;

    /// Network layer protocol used to ping, IPv4 or IPv6
    icmp::socket::protocol_type Protocol;

    /// The buffer where the data received will be placed
    boost::asio::streambuf ReplyBuffer;

    std::set<PingerItem> PingerList;

    /// Instances, one for each (protocol, interface) - pair
    static std::map<DistributorInstanceIdentifier, IcmpPaketDistributorItem,
                    IcmpPaketDistributorInstanceIdentifierComparator> Instances;

};


#endif  // ICMP_PAKET_DISTRIBUTOR_H

// (created using vim -- the world's best text editor)

